import os
import pwd
from twisted.python import failure
from twisted.internet import defer
from twisted.internet.endpoints import TCP4ClientEndpoint
from twisted.conch.ssh import connection
from carnifex.inductor import ProcessInductor
from carnifex.ssh.client import SSHClientFactory
from carnifex.ssh.command import SSHCommand
from carnifex.ssh.userauth import AutomaticUserAuthClient
from carnifex.ssh.session import execSession


class UnknownHostKey(Exception):
    """Raised when we reject validation of a server's host key
    """
    pass


class AllHostKeys(object):
    """Placeholder for known hosts container to allow all host keys.
    """
    def __contains__(self, key):
        return True
allHostKeys = AllHostKeys()


class SSHProcessInductor(ProcessInductor):
    _sep = ';' # '&&' to fail fast or ';' to continue on fails
    _cd = 'cd'
    _defaultUser = None
    knownHosts = allHostKeys

    def __init__(self, reactor, host, port, timeout=30, bindAddress=None,
                 precursor=None, credentials=None):
        self._connections = {}
        self.reactor = reactor
        self.endpoint = TCP4ClientEndpoint(reactor, host, port, timeout,
                                           bindAddress)
        self.precursor = precursor # 'source /etc/profile'
        self.credentials = credentials or {}

    def setCredentials(self, uid, password=None, privateKey=None, publicKey=None):
        user = self._getUser(uid)
        credentials = self.credentials.setdefault(user, {})
        credentials['password'] = password
        credentials['privateKey'] = privateKey
        credentials['publicKey'] = publicKey

    def addKnownHost(self, hostKey, fingerprint):
        if self.knownHosts == allHostKeys:
            self.knownHosts = {}
        self.knownHosts[fingerprint] = hostKey

    def execute(self, processProtocol, command, env={},
                path=None, uid=None, gid=None, usePTY=0, childFDs=None):
        """Execute a process on the remote machine using SSH

        @param processProtocol: the ProcessProtocol instance to connect
        @param executable: the executable program to run
        @param args: the arguments to pass to the process
        @param env: environment variables to request the remote ssh server to set
        @param path: the remote path to start the remote process on
        @param uid: user id or username to connect to the ssh server with
        @param gid: this is not used for remote ssh processes
        @param usePTY: wither to request a pty for the process
        @param childFDs: file descriptors to use for stdin, stdout and stderr
        """

        sshCommand = (command if isinstance(command, SSHCommand)
                      else SSHCommand(command, self.precursor, path))
        commandLine = sshCommand.getCommandLine()

        user = self._getUser(uid)

        # Get connection to ssh server
        connectionDeferred = self.getConnection(user)
        connectionDeferred.addCallback(execSession, processProtocol,
                                       commandLine, env, usePTY)
        return connectionDeferred

    def getConnection(self, user):
        #TODO: Fix case where we try to get another connection to the same user
        # before the first has connected...
        connection = self._connections.get(user, None)
        if connection:
            return defer.succeed(connection)
        # This will be called back if we already are in the process of connecting
        self._connections[user] = failure.Failure(Exception("Already trying to connect"))
        return self.startConnection(user)

    def startConnection(self, user):
        serviceStartedDeferred = defer.Deferred()
        connectionService = connection.SSHConnection()
        def serviceStarted():
            self._connections[user] = connectionService
            serviceStartedDeferred.callback(connectionService)
        connectionService.serviceStarted = serviceStarted

        connectionLostDeferred = defer.Deferred()
        userAuthObject = self._getUserAuthObject(user, connectionService)
        sshClientFactory = SSHClientFactory(connectionLostDeferred,
                                            self._verifyHostKey, userAuthObject)
        @connectionLostDeferred.addBoth
        def connectionLost(reason):
            serviceStartedDeferred.called or serviceStartedDeferred.errback(reason)
            self._connections[user] = None

        self.endpoint.connect(sshClientFactory)
        return serviceStartedDeferred

    def disconnectAll(self):
        for connection in self._connections.values():
            if hasattr(connection, 'transport'):
                connection.transport.loseConnection()

    def _getUser(self, uid):
        # Get the username from a uid, or use current user
        uid = uid or self._defaultUser or os.getuid()
        if isinstance(uid, int):
            user = pwd.getpwuid(uid).pw_name
        else:
            user = uid
        return user

    def _getCredentials(self, user):
        return self.credentials.get(user, dict(password=None,
                                               privateKey=None,
                                               publicKey=None))

    def _getUserAuthObject(self, user, connection):
        """Get a SSHUserAuthClient object to use for authentication

        @param user: The username to authenticate for
        @param connection: The connection service to start after authentication
        """
        credentials = self._getCredentials(user)
        userAuthObject = AutomaticUserAuthClient(user, connection, **credentials)
        return userAuthObject

    def _verifyHostKey(self, hostKey, fingerprint):
        """Called when ssh transport requests us to verify a given host key.
        Return a deferred that callback if we accept the key or errback if we
        decide to reject it.
        """
        if fingerprint in self.knownHosts:
            return defer.succeed(True)
        return defer.fail(UnknownHostKey(hostKey, fingerprint))

    def __del__(self):
        self.disconnectAll()



from twisted.internet._baseprocess import BaseProcess
from carnifex.inductor import ProcessInductor
from twisted.internet.error import ProcessTerminated, ProcessDone


class MockProcess(BaseProcess):
    def run(self, fauxProcessData):
        for childFd, data in fauxProcessData:
            self.proto.childDataReceived(childFd, data)

    def terminate(self, signal):
        """Simulate that the process was terminated with a signal
        """
        self.processEnded((None, signal))

    def exit(self, exitCode):
        """Simulate that the process exited
        """
        self.processEnded((exitCode, None))

    def _getReason(self, status):
        exitCode, signal = status
        if exitCode or signal:
            return ProcessTerminated(exitCode, signal, status)
        return ProcessDone(status)


class MockProcessInductor(ProcessInductor):
    def __init__(self, reactor, fauxProcessData, exitCode=0):
        self.reactor = reactor
        self.fauxProcessData = fauxProcessData
        self.exitCode = exitCode

    def execute(self, processProtocol, executable, args=(), env={},
                path=None, uid=None, gid=None, usePTY=0, childFDs=None):
        process = MockProcess(processProtocol)
        process.run(self.fauxProcessData)
        processProtocol.makeConnection(process)
        process.exit(self.exitCode)
        return process



from fabric.api import local

def unittest():
    local("python -m unittest discover -s test/unit")

def integrationtest():
    local("python -m unittest discover -s test/integration")

def test():
    local("python -m unittest discover")
    local("python setup.py check")

def release():
    test()
    local("python setup.py sdist upload")

def clean():
    local("python setup.py clean")

def whatsnew():
    local("git fetch")

    locallog = local("git log --abbrev-commit "
                     "--format='%Cgreen* %C(yellow)%h %Cblue%aN %Cgreen%ar "
                     "%Creset%s' FETCH_HEAD..", capture=True)
    remotelog = local("git log --abbrev-commit "
                      "--format='%Cred* %C(yellow)%h %Cblue%aN %Cgreen%ar "
                      "%Creset%s' ..FETCH_HEAD", capture=True)

    if locallog:
        print
        print "YOUR CHANGES:"
        print "-------------"
        print locallog
    if remotelog:
        print
        print "REMOTE CHANGES:"
        print "---------------"
        print remotelog

def push():
    test()
    local('git push')



from twisted.internet import defer
from twisted.internet.protocol import ProcessProtocol


class ProcessInductor(object):
    """Creates and follow up processes of local or remotely executed commands.
    """
    def execute(self, processProtocol, command, env={},
                path=None, uid=None, gid=None, usePTY=0, childFDs=None):
        """Form a command and start a process in the desired environment.
        """
        raise NotImplementedError()

    def run(self, command, env={}, path=None,
            uid=None, gid=None, usePTY=0, childFDs=None):
        """Execute a command and return the results of the completed run.
        """
        deferred = defer.Deferred()
        processProtocol = _SummaryProcessProtocol(deferred)
        d = defer.maybeDeferred(self.execute, processProtocol, command, env,
                     path, uid, gid, usePTY, childFDs)
        d.addErrback(deferred.errback)
        return deferred

    def getOutput(self, command, env={}, path=None,
                  uid=None, gid=None, usePTY=0, childFDs=None):
        """Execute a command and get the output of the finished process.
        """
        deferred = defer.Deferred()
        processProtocol = _SummaryProcessProtocol(deferred)
        self.execute(processProtocol, command, env,
                     path, uid, gid, usePTY, childFDs)
        @deferred.addCallback
        def getStdOut(tuple_):
            stdout, _stderr, _returnCode = tuple_
            return stdout
        return deferred

    def getExitCode(self, command, env={}, path=None, uid=None, gid=None,
                    usePTY=0, childFDs=None):
        """Execute a command and get the return code of the finished process.
        """
        deferred = defer.Deferred()
        processProtocol = _SummaryProcessProtocol(deferred)
        self.execute(processProtocol, command, env, path, uid, gid,
                     usePTY, childFDs)
        @deferred.addCallback
        def getStdOut(tuple_):
            _stdout, _stderr, exitCode = tuple_
            return exitCode
        return deferred


class _SummaryProcessProtocol(ProcessProtocol):
    """Gathers data of the process and delivers a summary when it completes.
    """

    def __init__(self, deferred, stdout=True, stderr=True, exitCode=True,
                 mergeErr=False):
        """
        @param deferred: the deferred to callback with the summary
        @param stdout: weither to gather data from stdout
        @param stderr: weither to gather data from stderr
        @param returnCode: weither to collect the return code (exit status)
        @param mergeErr: if True, merge stderr into stdout
        """
        self.deferred = deferred
        if stdout:
            self.bufOut = []
            self.outReceived = lambda data: self.bufOut.append(data)
        if mergeErr:
            self.errReceived = self.outReceived
        elif stderr:
            self.bufErr = []
            self.errReceived = lambda data: self.bufErr.append(data)

    def processEnded(self, reason):
        stdout = ''.join(getattr(self, 'bufOut', []))
        stderr = ''.join(getattr(self, 'bufErr', []))
        exitCode = reason.value.exitCode
        result = (stdout, stderr, exitCode)
        self.deferred.callback(result)



from twisted.trial.unittest import TestCase
from carnifex.localprocess import LocalProcessInductor
from twisted.internet.error import ConnectionDone
from twisted.internet import protocol, defer
from carnifex.endpoint import InductorEndpoint
from carnifex.sshprocess import SSHProcessInductor
from getpass import getpass


DEBUG = False
if DEBUG:
    import sys
    from twisted.python import log
    log.startLogging(sys.stdout)
    defer.Deferred.debug = True

UID = None # indicate that we want to run processes as the current user.
PASSWORD = None or getpass() # Set password here, or we will launch a prompt

ECHO_COMMAND = 'echo' # 'echo` should echo back on stdout
# `echo` return an additional newline after the given text to echo.
echoTransform = lambda text: text + '\n'


class GatherProtocol(protocol.Protocol):
    def __init__(self):
        self.resultDeferred = defer.Deferred()
        self.gatheredData = []

    def dataReceived (self, data):
        self.gatheredData.append(data)

    def connectionLost (self, reason):
        self.resultDeferred.callback((self.gatheredData, reason))


class InductorEndpointTestMixin(object):
    """Test connecting a twisted protocol to a process.
    """
    def test_real_endpoint(self):
        echoText = "hello world!"
        expectedGatheredData = [echoTransform(echoText)]
        expectedConnectionLostReason = ConnectionDone

        command = "%s %r" % (ECHO_COMMAND, echoText)
        endpoint = InductorEndpoint(self.inductor, command=command, timeout=1)

        protocolFactory = protocol.ClientFactory()
        protocolFactory.protocol = GatherProtocol
        clientDeferred = endpoint.connect(protocolFactory)

        @clientDeferred.addCallback
        def checkClient(client):
            self.assertIsInstance(client, GatherProtocol)
            return client.resultDeferred

        @clientDeferred.addCallback
        def checkGatheredData((gatheredData, connectionLostReason)):
            self.assertEqual(gatheredData, expectedGatheredData)
            return connectionLostReason

        # Check that the connection was lost as expected
        return self.assertFailure(clientDeferred, expectedConnectionLostReason)

        return clientDeferred


class LocalProcessInductorEndpointTest(TestCase, InductorEndpointTestMixin):
    def setUp(self):
        from twisted.internet import reactor
        self.inductor = LocalProcessInductor(reactor)


class SSHInductorEndpointTest(TestCase, InductorEndpointTestMixin):
    @classmethod
    def setUpClass(cls):
        from getpass import getpass
        cls.password = PASSWORD

    def setUp(self):
        from twisted.internet import reactor
        host, port = 'localhost', 22
        self.inductor = SSHProcessInductor(reactor, host, port)
        self.inductor.setCredentials(UID, self.password)

    def tearDown(self):
        inductor = self.inductor
        self.inductor = None
        inductor.disconnectAll()



from distutils.core import setup

README = open('README', 'r').read()

setup(
      name='carnifex',
      version='0.2.0',
      packages=['carnifex'],
      provides=['carnifex'],
      requires=['Twisted'],

      author='Geir Sporsheim',
      author_email='gksporsh@gmail.com',
      url='http://github.com/sporsh/carnifex',

      description="Execute commands and run processes locally and remote",
      classifiers=[
                   'Programming Language :: Python',
                   'Development Status :: 3 - Alpha',
                   'Framework :: Twisted',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Information Technology',
                   'Intended Audience :: System Administrators',
                   'Operating System :: OS Independent',
                   'Topic :: Internet',
                   'Topic :: Security',
                   'Topic :: System',
                   'Topic :: System :: Networking',
                   'Topic :: System :: Operating System',
                   'Topic :: System :: Systems Administration',
                   'Topic :: Software Development :: Libraries :: Python Modules',
                   'Topic :: Terminals'
                   ],
      long_description=README
      )



from twisted.internet.endpoints import _WrappingFactory
from twisted.internet.base import BaseConnector
from twisted.internet.protocol import ProcessProtocol, connectionDone
from twisted.internet import defer
from twisted.python.failure import Failure


class InductorEndpoint(object):
    """Endpoint that connects to a process spawned by an inductor.
    """
    def __init__(self, inductor, command, env={}, path=None,
                 uid=None, gid=None, usePTY=0, childFDs=None, timeout=None):
        self.inductor = inductor
        self.timeout = timeout
        self.inductorArgs = (command, env, path, uid, gid, usePTY, childFDs)

    def connect(self, protocolFactory):
        """Starts a process and connect a protocol to it.
        """
        deferred = self._startProcess()
        deferred.addCallback(self._connectRelay, protocolFactory)
        deferred.addCallback(self._startRelay)
        return deferred

    def _startProcess(self):
        """Use the inductor to start the process we want to relay data from.
        """
        connectedDeferred = defer.Deferred()
        processProtocol = RelayProcessProtocol(connectedDeferred)
        self.inductor.execute(processProtocol, *self.inductorArgs)
        return connectedDeferred

    def _connectRelay(self, process, protocolFactory):
        """Set up and connect the protocol we want to relay to the process.
        This method is automatically called when the process is started,
        and we are ready to relay through it.
        """
        try:
            wf = _WrappingFactory(protocolFactory)
            connector = RelayConnector(process, wf, self.timeout,
                                       self.inductor.reactor)
            connector.connect()
        except:
            return defer.fail()
        # Return a deferred that is called back when the protocol is connected.
        return wf._onConnection

    def _startRelay(self, client):
        """Start relaying data between the process and the protocol.
        This method is called when the protocol is connected.
        """
        process = client.transport.connector.process
        # Relay any buffered data that was received from the process before
        # we got connected and started relaying.
        for _, data in process.data:
            client.dataReceived(data)
        process.protocol = client

        @process._endedDeferred.addBoth
        def stopRelay(reason):
            """Stop relaying data. Called when the process has ended.
            """
            relay = client.transport
            relay.loseConnection(reason)
            connector = relay.connector
            connector.connectionLost(reason)

        # Pass through the client protocol.
        return client


class RelayTransport(object):
    """A protocol transport that relays to another transport.
    """
    connected = False
    disconnected = False

    def __init__(self, connector, reactor):
        self.connector = connector
        reactor.callLater(0, self.connectRelay)

    def connectRelay(self):
        """Builds the target protocol and connects it to the relay transport.
        """
        self.protocol = self.connector.buildProtocol(None)
        self.connected = True
        self.protocol.makeConnection(self)

    def _getTransport(self):
        return getattr(self.connector.proces, 'transport')

    def write(self, data):
        transport = self._getTransport()
        transport.write(data)

    def writeSequence(self, data):
        transport = self._getTransport()
        transport.writeSequence(data)

    def loseConnection(self, reason=connectionDone):
        protocol = getattr(self, 'protocol', None)
        if protocol:
            del self.protocol
            protocol.connectionLost(reason)
        self.disconnected = True

    def failIfNotConnected(self, err):
        if (self.connected or self.disconnected or
            not hasattr(self, "connector")):
            return

        self.connector.connectionFailed(Failure(err))
        del self.connector


class RelayConnector(BaseConnector):
    """Connect a protocol to a relaying transport.
    """
    def __init__(self, process, factory, timeout, reactor):
        if timeout:
            assert not reactor is None, "Need an reactor to use timeout."
        BaseConnector.__init__(self, factory, timeout, reactor)
        self.process = process

    def _makeTransport(self):
        relay = RelayTransport(self, self.reactor)
        self.process.relay = relay
        return relay


class RelayProcessProtocol(ProcessProtocol):
    """A process protocol that relays to a regular protocol.
    """
    def __init__(self, connectedDeferred):
        self._connectedDeferred = connectedDeferred
        self._endedDeferred = defer.Deferred()
        self.data = []

    def connectionMade(self):
        # We're connected to the process, so fire a deferred to connect the
        # target protocol and start relaying.
        self._connectedDeferred.callback(self)

    def childDataReceived(self, childFD, data):
        """Relay data received on any file descriptor to the process
        """
        protocol = getattr(self, 'protocol', None)
        if protocol:
            protocol.dataReceived(data)
        else:
            self.data.append((childFD, data))

    def processEnded(self, reason):
        # The process has ended, so we need to stop relaying.
        self._endedDeferred.callback(reason)
